# SETUP FOR ISING MODEL.  Stores information in global variables for later use.

ising_setup <- function (n_rows, n_cols)
{ n_rows <<- n_rows; n_cols <<- n_cols; n_spins <<- n_rows * n_cols; n_neighbors <<- 4
  neighbor <<- matrix (NA, n_spins, n_neighbors)
  for (j in 1:n_spins)
  { neighbor[j,1] <<- if (j%%n_rows==0) j-(n_rows-1) else j+1
    neighbor[j,2] <<- if (j%%n_rows==1) j+(n_rows-1) else j-1
    neighbor[j,3] <<- if ((j-1)%/%n_rows==n_cols-1) j-n_rows*(n_cols-1) else j+n_rows
    neighbor[j,4] <<- if ((j-1)%/%n_rows==0) j+n_rows*(n_cols-1) else j-n_rows
  }
}

# VECTORIZED MCMC FOR ISING MODEL.  Method is "s" for standard simulation with multiple
# random streams, "c" for coupled simulation with a single stream, and "p" for permuation
# MCMC with a single stream.  Initial values for states, u, and yf (y divided by pi) can be
# specified, or left to be generated randomly.  The "random" numbers can be supplied via the
# s argument (of dimensions n_iters by n_spins), and are otherwise set here using runif.  
# If the rev argument is TRUE, the reverse map is applied.  The value returned is an array 
# of states at all iterations, with attributes giving end values of states, u, and yf.

vec_ising <- function (beta, method, n_chains, n_iters, states=NULL, u=NULL, yf=NULL, 
                       s=NULL, rev=FALSE)
{ if (is.null(states))
    states <- matrix (sample (c(-1,1), n_chains*n_spins, replace=TRUE), n_spins, n_chains)
  if (is.null(u) && method=="p") u <- runif (n_chains)
  if (is.null(yf) && method=="p") yf <- runif (n_chains) 
  if (is.null(s) && (method=="p" || method=="c"))
    s <- matrix(runif(n_iters*n_spins),n_iters,n_spins)

  if (rev) s <- s[n_iters:1,,drop=FALSE]

  results <- array(dim=c(n_spins,n_chains,n_iters))

  for (i in 1:n_iters)
  { for (j in { if (rev) n_spins:1 else 1:n_spins })
    { 
      if (method=="p" && rev) u <- (u - s[i,j]) %% 1    # Take s from u first for reverse map

      neighbor_sum <- rep(0,n_chains)
      for (k in 1:n_neighbors) neighbor_sum <- neighbor_sum + states[neighbor[j,k],]

      if (method=="s") u <- runif(n_chains)
      if (method=="c") u <- s[i,j]
      if (method=="p") old_spin01 <- states[j,] == 1    # Old values of spins in 0/1 form

      p0 <- 1 / (1 + exp(2*beta*neighbor_sum))          # Probabilities of setting spin to 0
      spin01 <- p0 <= u                                 # New values of spins in 0/1 form
      states[j,] <- 2*spin01 - 1                        # New values of spins in -1/+1 form

      if (method=="p")
      { t <- spin01*(1-p0) + (1-spin01)*p0              # Probabilities of new spin values
        old_t <- old_spin01*(1-p0) + (1-old_spin01)*p0  # Probabilities of old spin values 
        old_yf <- yf
        yf <- (u - p0*spin01) / t                       # New yf (equal to y/pi) from eq (23)
        u <- p0*old_spin01 + old_t*old_yf               # New u (except for + s) from eq (24) 
        if (!rev) u <- (u + s[i,j]) %% 1                # Add s to u last for non-reverse map
      }
    }
    results[,,i] <- states
  }

  structure (results, end_states=states, end_u=u, end_yf=yf)
}

# COMPUTE THE ENERGY OF A SPIN CONFIGURATION.

ising_energy <- function (spins)
{ energy <- 0
  for (j in 1:n_spins) energy <- energy - spins[j] * sum(spins[neighbor[j,]])
  energy / 2  # Divide by two because each term in the energy was added twice above
}

# PLOT THE ENERGY ALONG ALL CHAINS.

energy_plot <- function (res)
{ n_chains <- dim(res)[2]; n_iters <- dim(res)[3]

  E <- matrix(NA,n_iters,n_chains)
  for (i in 1:n_iters) for (k in 1:n_chains) E[i,k] <- ising_energy(res[,k,i])
  E_bar <- colMeans(E)

  cat ("Energy: mean ",round(mean(E_bar),3), " +- ",round(sd(E_bar)/sqrt(n_chains),3),
       "\n",sep="")

  plot (c(1,n_iters), c(-2*n_spins, 2*n_spins), type="n", xlab="Iteration", ylab="")
  for (k in 1:n_chains) points (E[,k], col=k, pch=20)
}

# PLOT THE MAGNETIZATION (SUM OF SPINS) ALONG ALL CHAINS.

magnetization_plot <- function (res)
{ n_chains <- dim(res)[2]; n_iters <- dim(res)[3]

  M <- matrix(NA,n_iters,n_chains)
  for (i in 1:n_iters) for (k in 1:n_chains) M[i,k] <- sum(res[,k,i])
  M_bar <- colMeans(M); M_abs_bar <- colMeans(abs(M))

  cat ("Magnetization, mean ",round(mean(M_bar),3), " +- ",round(sd(M_bar)/sqrt(n_chains),3),
       ", mean abs ",round(mean(M_abs_bar),3), " +- ",round(sd(M_abs_bar)/sqrt(n_chains),3),
       "\n",sep="")

  plot (c(1,n_iters), c(-n_spins, n_spins), type="n", xlab="Iteration", ylab="")
  for (k in 1:n_chains) points (apply(res[,k,],2,sum), col=k, pch=20)
}
