# LOG PROBABILITY DENSITY FUNCTION FOR TEST DISTRIBUTION.

istest_log_density <- function (x) dnorm(x[1],0,1,log=TRUE) + dnorm(x[2],x[1]^2-1,1,log=TRUE)

# VECTORIZED PERMUTATION METROPOLIS FOR TEST DISTRIBUTION. Does random-walk Metropolis updates
# for both variables at once.  If delta is not supplied, it's normal with given proposal_sd.

vec_istest_met <- function (n_chains, n_iters, states=NULL, u=NULL, yf=NULL, 
                            s=NULL, delta=NULL, rev=FALSE, proposal_sd)
{ 
  if (is.null(states)) states <- matrix (rnorm(n_chains*2,0,1), 2, n_chains)
  if (is.null(u))      u <- runif(n_chains)
  if (is.null(yf))     yf <- runif(n_chains)
  if (is.null(s))      s <- runif(n_iters)
  if (is.null(delta))  delta <- matrix(rnorm(n_iters*2,0,proposal_sd),n_iters,2)
  if (rev) { s <- s[n_iters:1]; delta <- delta[n_iters:1,,drop=FALSE] }

  results <- array(dim=c(2,n_chains,n_iters))

  for (i in 1:n_iters)
  { 
    if (rev) u <- (u - s[i]) %% 1

    proposal <- t (t(states+delta[i,]) * (u<1/2) + t(states-delta[i,]) * (u>=1/2))
    a <- (2*u) %% 1
    ar <- exp (apply(proposal,2,istest_log_density) - apply(states,2,istest_log_density))
    ap <- pmin(1,ar)
    accept <- (a < ap)
    states <- t (accept*t(proposal) + (1-accept)*t(states))
    u[accept] <- ((u<1/2)/2 + pmin(1,1/ar)*yf/2)[accept]
    yf[accept] <- (a/ap)[accept]

    if (!rev) u <- (u + s[i]) %% 1

    results[,,i] <- states
  }

  structure (results, end_states=states, end_u=u, end_yf=yf)
}

# METROPOLIS IMPORTANCE SAMPLING FOR TEST DISTRIBUTION.  N is number of points, M is the
# maximum number of transitons, mu and sd are for the normal start state distribution, and 
# proposal_sd is for the normal random-walk Metropolis proposals.

istest_met_is <- function (N, M, mu, sd, proposal_sd)
{
  pr <- rep(NA,M+1); w <- rep(NA,N); beta <- matrix (NA,N,2); states <- matrix(NA,2,M+1)
  s <- runif(M); delta <- matrix (rnorm(M*2,0,proposal_sd), M, 2)
  
  for (i in 1:N)
  { 
    k <- sample(0:M,1)                   # Pick number of transitions from start to end state

    states[,k+1] <- rnorm(2,mu,sd)       # Random start state
    u <- runif(1); yf <- runif(1)        # Random start values for u and yf

    if (k<M)                             # Do a forward simulation from the start state on
      states[,(k+2):(M+1)] <- 
        vec_istest_met (1, M-k, cbind(states[,k+1]), u, yf,
                        s=s[(k+1):M], delta=delta[(k+1):M,,drop=FALSE], rev=FALSE) [,1,]

    if (k>0)                             # Do a reverse simulation before the start state
      states[,1:k] <- 
        vec_istest_met (1, k, cbind(states[,k+1]), u, yf,
                        s=s[1:k], delta=delta[1:k,,drop=FALSE], rev=TRUE) [,1,]

    beta[i,] <- states[,M+1]             # Importance sampling point is the end state

    for (k in 0:M)                       
      pr[k+1] <- sum(dnorm(states[,k+1],mu,sd,log=TRUE)) -      # Compute log probabilities 
                     istest_log_density(states[,k+1])           #   of possible start states
    w[i] <- - (max(pr) + log(sum(exp(pr-max(pr)))) - log(M+1))  # Log importance weight =
  }                                                             #   - log average probability

  list (w=w, beta=beta)
}

# ANALYSE IMPORTANCE SAMPLING OUTPUT.  Takes as arguments a vector of log weights and a
# matrix of beta values.  Prints various estimates with standard errors.

is_analysis <- function (w, beta)
{
  N <- length(w)
  mx <- max(w)
  w <- exp (w-mx)

  cat ("\nEstimated log normalizing constant:", round(mx+log(mean(w)),3), 
       "+-", round(sd(w)/sqrt(N)/mean(w),3), "\n\n")

  w <- w / mean(w)

  z <- sum(w)
  ave <- colSums(w*beta) / z
  ave_sq <- colSums(w*beta^2) / z

  cat ("Adjusted sample size:", N, "/", 1+var(w), "=", N/(1+var(w)), "\n\n")

  beta_offset <- t (t(beta) - ave)
  ave_se <- sqrt (colSums((w*beta_offset)^2) / N^2)
  beta_sq_offset <- t (t(beta^2) - ave_sq)
  ave_sq_se <- sqrt (colSums((w*beta_sq_offset)^2) / N^2)

  print(cbind (ave=ave, se=ave_se, ave_sq=ave_sq, sq_se=ave_sq_se))
}

