# VECTORIZED GIBSS SAMPLING FOR TRUNCATED NORMAL MODEL.  Method is "s" for standard simulation
# "c" for coupled simulation, and "p" for permuation MCMC.  Initial values and the random 
# stream can be specified, or left to be generated randomly.  If the rev argument is TRUE, 
# the map is reversed.  The value returned is an array of states at all iterations, with 
# attributes giving end values of states and u.  Uses global rho, lower, and upper variables.

vec_trunc_normal_gibbs <- function (method, n_chains, n_iters, states=NULL, u=NULL,
                                    s=NULL, rev=FALSE) 
{ if (is.null(states)) 
  { states <- matrix (NA, 2, n_chains)
    for (j in 1:2) states[j,] <- runif(n_chains, lower[j], upper[j])
  }
  if (is.null(u) && method=="p") u <- runif (n_chains)
  if (is.null(s) && (method=="p" || method=="c")) s <- matrix(runif(2*n_iters),n_iters,2)
  if (rev) s <- s[n_iters:1,,drop=FALSE]

  results <- array(dim=c(2,n_chains,n_iters))

  for (i in 1:n_iters)
  { for (j in { if (rev) 2:1 else 1:2 })
    { if (method=="p" && rev) u <- (u - s[i,j]) %% 1
      if (method=="s") u <- runif(n_chains)
      if (method=="c") u <- s[i,j]
      if (method=="p") old <- states[j,]
      cond_sd <- sqrt(1-rho^2); cond_mean <- rho * states[3-j,]
      ql <- pnorm (lower[j], cond_mean, cond_sd); qu <- pnorm (upper[j], cond_mean, cond_sd)
      states[j,] <- # max & min below not usually needed - just in case of numerical issue
        pmax (lower[j], pmin (upper[j], qnorm (ql + (qu-ql)*u, cond_mean, cond_sd)))
      if (method=="p") 
      { u <- (pnorm(old,cond_mean,cond_sd) - ql) / (qu-ql)
        if (!rev) u <- (u + s[i,j]) %% 1
      }
    }
    results[,,i] <- states
  }
  structure (results, end_states=states, end_u=u)
}

# VECTORIZED METROPOLIS SAMPLING FOR TRUNCATED NORMAL MODEL.  Arguments are as above, but 
# with no "rev" argument, an extra optional "yf" argument, and an extra optional "delta" 
# argument that specifies random-walk proposal offsets for each variable update.

vec_trunc_normal_met <- function (method, n_chains, n_iters, states=NULL, u=NULL, yf=NULL,
                                  s=NULL, delta=NULL)
{ if (is.null(states)) 
  { states <- matrix (NA, 2, n_chains)
    for (j in 1:2) states[j,] <- runif(n_chains, lower[j], upper[j])
  }
  if (is.null(u) && method=="p") u <- runif (n_chains)
  if (is.null(yf) && method=="p") yf <- runif(n_chains)
  if (is.null(s) && (method=="p" || method=="c")) s <- matrix(runif(2*n_iters),n_iters,2)
  if (is.null(delta)) delta <- matrix(runif(2*n_iters,0,1),n_iters,2)

  results <- array(dim=c(2,n_chains,n_iters))

  for (i in 1:n_iters)
  { for (j in 1:2)
    { if (method=="s") u <- runif(n_chains)
      if (method=="c") u <- s[i,j]
      old <- states[j,]
      cond_sd <- sqrt(1-rho^2); cond_mean <- rho * states[3-j,]
      proposal <- (states[j,]+delta[i,j]) * (u<1/2) + (states[j,]-delta[i,j]) * (u>=1/2)
      a <- (2*u) %% 1
      ar <- dnorm(proposal,cond_mean,cond_sd) / dnorm(old,cond_mean,cond_sd)
      ap <- (proposal >= lower[j]) * (proposal <= upper[j]) * pmin(1,ar)
      accept <- (a < ap)
      states[j,] <- accept*proposal + (1-accept)*old
      if (method=="p") 
      { u[accept] <- ((u<1/2)/2 + pmin(1,1/ar)*yf/2)[accept]; yf[accept] <- (a/ap)[accept]
        u <- (u + s[i,j]) %% 1
      }
    }
    results[,,i] <- states
  }
  results
}

# PLOT X1 VERSUS X2 FOR SIMULATIONS. 

bi_plot <- function (res, from=11, xlim=c(lower[1],upper[1]), ylim=c(lower[2],upper[2]))
{ n_chains <- dim(res)[2]; n_iters <= dim(res)[3]
  plot (c(), type="n", xlab="x1", ylab="x2", xlim=xlim, ylim=ylim)
  for (k in 1:n_chains) points (res[1,k,from:n_iters], res[2,k,from:n_iters], pch=20, col=k)
}

# PLOT X1 OR X2 VERSUS ITERATION NUMBER FOR SIMULATIONS.

uni_plot <- function (res, j, ylim=c(lower[j],upper[j]))
{ n_chains <- dim(res)[2]; n_iters <- dim(res)[3]
  plot (c(),type="n",xlab="Iteration",ylab=paste("x",j,sep=""),xlim=c(1,n_iters),ylim=ylim)
  for (k in 1:n_chains) points (res[j,k,], pch=20, col=k)
}

# FIND MEAN ESTIMATES AND STANDARD ERRORS FROM SIMULATIONS.  

estimates <- function (res, from=11)
{ n_chains <- dim(res)[2]; n_iters <- dim(res)[3]
  ave <- ave_sq <- matrix(NA,2,n_chains)
  for (k in 1:n_chains) 
  { ave[,k] <- rowMeans(res[,k,from:n_iters]);
    ave_sq[,k] <- rowMeans(res[,k,from:n_iters]^2) 
  }
  round (cbind (ave=apply(ave,1,mean),       se=apply(ave,1,sd)/sqrt(n_chains), 
                ave_sq=apply(ave_sq,1,mean), se_sq=apply(ave_sq,1,sd)/sqrt(n_chains)), 4)
}
